from .annotation import Annotation
from .feature import Feature
from .gene import Gene
from .genome import Genome, Scaffold
from .transcript import Transcript
from .subfeatures import CDS, Exon, UTR, Intron
from .misc_features import Protein, Promoter
from .hits import OverlapHit, BlastHit